#!/opt/venvs/paasta-tools/bin/python
import os
import sys
from filecmp import cmp
from shutil import copy
from subprocess import CalledProcessError
from subprocess import run
from traceback import print_exc

APPLIED_DIRECTORY = ".applied"


# This script expects the KUBECONFIG environment variable to be set correctly
def main(puppet_resource_root: str) -> int:
    exit_code = 0
    applied_resource_root = os.path.join(puppet_resource_root, APPLIED_DIRECTORY)

    # Loop through everything in the puppet resource path
    for root, dirs, files in os.walk(puppet_resource_root):
        # modifying the 'dirs' variable in-place will update the order that os.walk visits things
        if APPLIED_DIRECTORY in dirs:
            dirs.remove(APPLIED_DIRECTORY)
        dirs.sort()  # Need to apply things in the correct order

        # Check to see if there's a difference between what Puppet created and
        # what's been previously applied
        for filename in sorted([f for f in files if f.endswith(".yaml")]):
            path = os.path.join(root, filename)
            applied_path = os.path.join(
                applied_resource_root, os.path.relpath(path, puppet_resource_root)
            )
            print(f"comparing {path} and {applied_path}")
            if not os.path.exists(applied_path) or not cmp(
                path, applied_path, shallow=False
            ):
                # This is idempotent; if something gets out of sync and a resource gets applied
                # a second time, kubectl just won't make any changes
                try:
                    run(["kubectl", "apply", "-f", path], check=True)
                    os.makedirs(os.path.dirname(applied_path), exist_ok=True)
                    copy(path, applied_path)
                except CalledProcessError:
                    print(f"There was a problem applying {path}:\n")
                    print_exc(
                        file=sys.stdout
                    )  # keep all messages on the same stream so they're in order
                    exit_code = 1
                    continue

    # Loop through all the files that have been previously applied and see
    # if Puppet has removed any of them
    for root, dirs, files in os.walk(applied_resource_root):
        dirs.sort(reverse=True)  # for deleting things, we need to go in reverse order
        for filename in sorted([f for f in files if f.endswith(".yaml")], reverse=True):
            path = os.path.join(root, filename)
            puppet_path = os.path.join(
                puppet_resource_root, os.path.relpath(path, applied_resource_root)
            )
            if not os.path.exists(puppet_path):
                print(f"Deleting resource {path}...")
                try:
                    run(
                        ["kubectl", "delete", "--ignore-not-found=true", "-f", path],
                        check=True,
                    )
                    os.remove(path)
                except CalledProcessError:
                    print(f"There was a problem deleting {path}:\n")
                    print_exc(
                        file=sys.stdout
                    )  # keep all messages on the same stream so they're in order
                    exit_code = 1
                    continue

    return exit_code


if __name__ == "__main__":
    sys.exit(main(sys.argv[1]))



from typing import Optional

import service_configuration_lib

from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfigDict
from paasta_tools.utils import BranchDictV2
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import load_service_instance_config
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import time_cache


class EksDeploymentConfig(KubernetesDeploymentConfig):
    config_dict: KubernetesDeploymentConfigDict

    config_filename_prefix = "eks"

    def __init__(
        self,
        service: str,
        cluster: str,
        instance: str,
        config_dict: KubernetesDeploymentConfigDict,
        branch_dict: Optional[BranchDictV2],
        soa_dir: str = DEFAULT_SOA_DIR,
    ) -> None:
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            config_dict=config_dict,
            branch_dict=branch_dict,
            soa_dir=soa_dir,
        )


def load_eks_service_config_no_cache(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> "EksDeploymentConfig":
    """Read a service instance's configuration for EKS.

    If a branch isn't specified for a config, the 'branch' key defaults to
    paasta-${cluster}.${instance}.

    :param name: The service name
    :param instance: The instance of the service to retrieve
    :param cluster: The cluster to read the configuration for
    :param load_deployments: A boolean indicating if the corresponding deployments.json for this service
                             should also be loaded
    :param soa_dir: The SOA configuration directory to read from
    :returns: A dictionary of whatever was in the config for the service instance"""
    general_config = service_configuration_lib.read_service_configuration(
        service, soa_dir=soa_dir
    )
    instance_config = load_service_instance_config(
        service, instance, "eks", cluster, soa_dir=soa_dir
    )
    general_config = deep_merge_dictionaries(
        overrides=instance_config, defaults=general_config
    )

    branch_dict: Optional[BranchDictV2] = None
    if load_deployments:
        deployments_json = load_v2_deployments_json(service, soa_dir=soa_dir)
        temp_instance_config = EksDeploymentConfig(
            service=service,
            cluster=cluster,
            instance=instance,
            config_dict=general_config,
            branch_dict=None,
            soa_dir=soa_dir,
        )
        branch = temp_instance_config.get_branch()
        deploy_group = temp_instance_config.get_deploy_group()
        branch_dict = deployments_json.get_branch_dict(service, branch, deploy_group)

    return EksDeploymentConfig(
        service=service,
        cluster=cluster,
        instance=instance,
        config_dict=general_config,
        branch_dict=branch_dict,
        soa_dir=soa_dir,
    )


@time_cache(ttl=5)
def load_eks_service_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> "EksDeploymentConfig":
    """Read a service instance's configuration for EKS.

    If a branch isn't specified for a config, the 'branch' key defaults to
    paasta-${cluster}.${instance}.

    :param name: The service name
    :param instance: The instance of the service to retrieve
    :param cluster: The cluster to read the configuration for
    :param load_deployments: A boolean indicating if the corresponding deployments.json for this service
                             should also be loaded
    :param soa_dir: The SOA configuration directory to read from
    :returns: A dictionary of whatever was in the config for the service instance"""
    return load_eks_service_config_no_cache(
        service=service,
        instance=instance,
        cluster=cluster,
        load_deployments=load_deployments,
        soa_dir=soa_dir,
    )



import json
import os
from typing import Dict
from typing import List
from typing import Tuple

import service_configuration_lib

from paasta_tools.long_running_service_tools import load_service_namespace_config
from paasta_tools.long_running_service_tools import ServiceNamespaceConfig
from paasta_tools.utils import compose_job_id

PUPPET_SERVICE_DIR = "/etc/nerve/puppet_services.d"


def get_puppet_services_that_run_here() -> Dict[str, List[str]]:
    # find all files in the PUPPET_SERVICE_DIR, but discard broken symlinks
    # this allows us to (de)register services on a machine by
    # breaking/healing a symlink placed by Puppet.
    puppet_service_dir_services = {}
    if os.path.exists(PUPPET_SERVICE_DIR):
        for service_name in os.listdir(PUPPET_SERVICE_DIR):
            if not os.path.exists(os.path.join(PUPPET_SERVICE_DIR, service_name)):
                continue
            with open(os.path.join(PUPPET_SERVICE_DIR, service_name)) as f:
                puppet_service_data = json.load(f)
                puppet_service_dir_services[service_name] = puppet_service_data[
                    "namespaces"
                ]

    return puppet_service_dir_services


def get_puppet_services_running_here_for_nerve(
    soa_dir: str,
) -> List[Tuple[str, ServiceNamespaceConfig]]:
    puppet_services = []
    for service, namespaces in sorted(get_puppet_services_that_run_here().items()):
        for namespace in namespaces:
            puppet_services.append(
                _namespaced_get_classic_service_information_for_nerve(
                    service, namespace, soa_dir
                )
            )
    return puppet_services


def _namespaced_get_classic_service_information_for_nerve(
    name: str, namespace: str, soa_dir: str
) -> Tuple[str, ServiceNamespaceConfig]:
    nerve_dict = load_service_namespace_config(name, namespace, soa_dir)
    port_file = os.path.join(soa_dir, name, "port")
    # If the namespace defines a port, prefer that, otherwise use the
    # service wide port file.
    nerve_dict["port"] = nerve_dict.get(
        "port", None
    ) or service_configuration_lib.read_port(port_file)
    nerve_name = compose_job_id(name, namespace)
    return (nerve_name, nerve_dict)



# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from typing import Dict
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Type

from service_configuration_lib import read_service_configuration

from paasta_tools import utils
from paasta_tools.utils import deep_merge_dictionaries
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig_T
from paasta_tools.utils import list_clusters
from paasta_tools.utils import load_service_instance_configs
from paasta_tools.utils import load_v2_deployments_json
from paasta_tools.utils import NoDeploymentsAvailable


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class PaastaServiceConfigLoader:
    """PaastaServiceConfigLoader provides useful methods for reading soa-configs and
    iterating instance names or InstanceConfigs objects.

    :Example:

    >>> from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
    >>> from paasta_tools.utils import DEFAULT_SOA_DIR
    >>>
    >>> sc = PaastaServiceConfigLoader(service='fake_service', soa_dir=DEFAULT_SOA_DIR)
    >>>
    >>> for instance in sc.instances(cluster='fake_cluster', instance_type_class=KubernetesDeploymentConfig):
    ...     print(instance)
    ...
    main
    canary
    >>>
    >>> for instance_config in sc.instance_configs(cluster='fake_cluster', instance_type_class=KubernetesDeploymentConfig):
    ...     print(instance_config.get_instance())
    ...
    main
    canary
    >>>
    """

    _framework_configs: Dict[Tuple[str, type], Dict[str, utils.InstanceConfigDict]]
    _clusters: List[str]
    _deployments_json: utils.DeploymentsJsonV2

    def __init__(
        self,
        service: str,
        soa_dir: str = DEFAULT_SOA_DIR,
        load_deployments: bool = True,
    ) -> None:
        self._service = service
        self._soa_dir = soa_dir
        self._load_deployments = load_deployments
        self._clusters = None
        self._general_config = None
        self._deployments_json = None
        self._framework_configs = {}

    @property
    def clusters(self) -> Iterable[str]:
        """Returns an iterator that yields cluster names for the service.

        :returns: iterator that yields cluster names.
        """
        if self._clusters is None:
            self._clusters = list_clusters(service=self._service, soa_dir=self._soa_dir)
        for cluster in self._clusters:
            yield cluster

    def instances(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ) -> Iterable[str]:
        """Returns an iterator that yields instance names as strings.

        :param cluster: The cluster name
        :param instance_type_class: a subclass of InstanceConfig
        :returns: an iterator that yields instance names
        """
        if (cluster, instance_type_class) not in self._framework_configs:
            self._refresh_framework_config(cluster, instance_type_class)
        for instance in self._framework_configs.get((cluster, instance_type_class), []):
            yield instance

    def instance_configs(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ) -> Iterable[InstanceConfig_T]:
        """Returns an iterator that yields InstanceConfig objects.

        :param cluster: The cluster name
        :param instance_type_class: a subclass of InstanceConfig
        :returns: an iterator that yields instances of KubernetesDeploymentConfig, etc.
        :raises NotImplementedError: when it doesn't know how to create a config for instance_type_class
        """
        if (cluster, instance_type_class) not in self._framework_configs:
            self._refresh_framework_config(cluster, instance_type_class)
        for instance, config in self._framework_configs.get(
            (cluster, instance_type_class), {}
        ).items():
            try:
                yield self._create_service_config(
                    cluster, instance, config, instance_type_class
                )
            except NoDeploymentsAvailable:
                pass

    def _framework_config_filename(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ):
        return f"{instance_type_class.config_filename_prefix}-{cluster}"

    def _refresh_framework_config(
        self, cluster: str, instance_type_class: Type[InstanceConfig_T]
    ):
        instances = load_service_instance_configs(
            service=self._service,
            instance_type=instance_type_class.config_filename_prefix,
            cluster=cluster,
            soa_dir=self._soa_dir,
        )
        self._framework_configs[(cluster, instance_type_class)] = instances

    def _get_branch_dict(
        self, cluster: str, instance: str, config: utils.InstanceConfig
    ) -> utils.BranchDictV2:
        if self._deployments_json is None:
            self._deployments_json = load_v2_deployments_json(
                self._service, soa_dir=self._soa_dir
            )

        branch = config.get_branch()
        deploy_group = config.get_deploy_group()
        return self._deployments_json.get_branch_dict(
            self._service, branch, deploy_group
        )

    def _get_merged_config(
        self, config: utils.InstanceConfigDict
    ) -> utils.InstanceConfigDict:
        if self._general_config is None:
            self._general_config = read_service_configuration(
                service_name=self._service, soa_dir=self._soa_dir
            )
        return deep_merge_dictionaries(overrides=config, defaults=self._general_config)

    def _create_service_config(
        self,
        cluster: str,
        instance: str,
        config: utils.InstanceConfigDict,
        config_class: Type[InstanceConfig_T],
    ) -> InstanceConfig_T:
        """Create a service instance's configuration for kubernetes.

        :param cluster: The cluster to read the configuration for
        :param instance: The instance of the service to retrieve
        :param config: the framework instance config.
        :returns: An instance of config_class
        """

        merged_config = self._get_merged_config(config)

        temp_instance_config = config_class(
            service=self._service,
            cluster=cluster,
            instance=instance,
            config_dict=merged_config,
            branch_dict=None,
            soa_dir=self._soa_dir,
        )

        branch_dict = self._get_branch_dict(cluster, instance, temp_instance_config)

        return config_class(
            service=self._service,
            cluster=cluster,
            instance=instance,
            config_dict=merged_config,
            branch_dict=branch_dict,
            soa_dir=self._soa_dir,
        )



# coding: utf-8

# flake8: noqa

# import all models into this package
# if you have many models here with many references from one model to another this may
# raise a RecursionError
# to avoid this, import only the models that you directly need like:
# from from paasta_tools.paastaapi.model.pet import Pet
# or import this package, but before doing it, use:
# import sys
# sys.setrecursionlimit(n)

from paasta_tools.paastaapi.model.adhoc_launch_history import AdhocLaunchHistory
from paasta_tools.paastaapi.model.autoscaler_count_msg import AutoscalerCountMsg
from paasta_tools.paastaapi.model.deploy_queue import DeployQueue
from paasta_tools.paastaapi.model.deploy_queue_service_instance import DeployQueueServiceInstance
from paasta_tools.paastaapi.model.envoy_backend import EnvoyBackend
from paasta_tools.paastaapi.model.envoy_location import EnvoyLocation
from paasta_tools.paastaapi.model.envoy_status import EnvoyStatus
from paasta_tools.paastaapi.model.flink_cluster_overview import FlinkClusterOverview
from paasta_tools.paastaapi.model.flink_config import FlinkConfig
from paasta_tools.paastaapi.model.flink_job import FlinkJob
from paasta_tools.paastaapi.model.flink_job_details import FlinkJobDetails
from paasta_tools.paastaapi.model.flink_jobs import FlinkJobs
from paasta_tools.paastaapi.model.float_and_error import FloatAndError
from paasta_tools.paastaapi.model.hpa_metric import HPAMetric
from paasta_tools.paastaapi.model.inline_object import InlineObject
from paasta_tools.paastaapi.model.inline_response200 import InlineResponse200
from paasta_tools.paastaapi.model.inline_response2001 import InlineResponse2001
from paasta_tools.paastaapi.model.instance_bounce_status import InstanceBounceStatus
from paasta_tools.paastaapi.model.instance_mesh_status import InstanceMeshStatus
from paasta_tools.paastaapi.model.instance_status import InstanceStatus
from paasta_tools.paastaapi.model.instance_status_adhoc import InstanceStatusAdhoc
from paasta_tools.paastaapi.model.instance_status_cassandracluster import InstanceStatusCassandracluster
from paasta_tools.paastaapi.model.instance_status_flink import InstanceStatusFlink
from paasta_tools.paastaapi.model.instance_status_kafkacluster import InstanceStatusKafkacluster
from paasta_tools.paastaapi.model.instance_status_kubernetes import InstanceStatusKubernetes
from paasta_tools.paastaapi.model.instance_status_kubernetes_autoscaling_status import InstanceStatusKubernetesAutoscalingStatus
from paasta_tools.paastaapi.model.instance_status_kubernetes_v2 import InstanceStatusKubernetesV2
from paasta_tools.paastaapi.model.instance_status_tron import InstanceStatusTron
from paasta_tools.paastaapi.model.instance_status_vitesscluster import InstanceStatusVitesscluster
from paasta_tools.paastaapi.model.instance_tasks import InstanceTasks
from paasta_tools.paastaapi.model.integer_and_error import IntegerAndError
from paasta_tools.paastaapi.model.kubernetes_container import KubernetesContainer
from paasta_tools.paastaapi.model.kubernetes_container_v2 import KubernetesContainerV2
from paasta_tools.paastaapi.model.kubernetes_healthcheck import KubernetesHealthcheck
from paasta_tools.paastaapi.model.kubernetes_pod import KubernetesPod
from paasta_tools.paastaapi.model.kubernetes_pod_event import KubernetesPodEvent
from paasta_tools.paastaapi.model.kubernetes_pod_v2 import KubernetesPodV2
from paasta_tools.paastaapi.model.kubernetes_replica_set import KubernetesReplicaSet
from paasta_tools.paastaapi.model.kubernetes_version import KubernetesVersion
from paasta_tools.paastaapi.model.resource import Resource
from paasta_tools.paastaapi.model.resource_item import ResourceItem
from paasta_tools.paastaapi.model.resource_value import ResourceValue
from paasta_tools.paastaapi.model.smartstack_backend import SmartstackBackend
from paasta_tools.paastaapi.model.smartstack_location import SmartstackLocation
from paasta_tools.paastaapi.model.smartstack_status import SmartstackStatus
from paasta_tools.paastaapi.model.task_tail_lines import TaskTailLines



import copy
import logging
import os
import subprocess
from tempfile import TemporaryDirectory
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import ruamel.yaml as yaml

from paasta_tools.cli.cmds.validate import validate_schema
from paasta_tools.utils import AUTO_SOACONFIG_SUBDIR
from paasta_tools.utils import DEFAULT_SOA_DIR


log = logging.getLogger(__name__)

# Must have a schema defined
KNOWN_CONFIG_TYPES = (
    "kubernetes",
    "deploy",
    "smartstack",
    "cassandracluster",
)

# this could use a better name - but basically, this is for pairs of instance types
# where you generally want to check both types (i.e.,g a status-quo and migration
# instance type)
INSTANCE_TYPE_COUNTERPARTS = {"eks": "kubernetes", "kubernetes": "eks"}


def my_represent_none(self, data):
    return self.represent_scalar("tag:yaml.org,2002:null", "null")


def write_auto_config_data(
    service: str,
    extra_info: str,
    data: Dict[str, Any],
    soa_dir: str = DEFAULT_SOA_DIR,
    sub_dir: Optional[str] = None,
    comment: Optional[str] = None,
) -> Optional[str]:
    """
    Replaces the contents of an automated config file for a service, or creates the file if it does not exist.

    Returns the filename of the modified file, or None if no file was written.
    """
    yaml.YAML().representer.add_representer(type(None), my_represent_none)
    service_dir = f"{soa_dir}/{service}"
    if not os.path.exists(service_dir):
        log.warning(
            f"Service {service} does not exist in configs, skipping auto config update"
        )
        return None
    subdir = f"{service_dir}/{sub_dir}" if sub_dir else service_dir
    if not os.path.exists(subdir):
        os.mkdir(subdir)
    filename = f"{subdir}/{extra_info}.yaml"

    with open(filename, "w") as f:
        # TODO: this can be collapsed into one codeblock. It is separated as two
        # because doing content.update(data) results in losing comments from `data`
        # we should be able to handle adding a header comment and yaml with comments in it
        # without this if/else block
        if comment:
            content = (
                yaml.round_trip_load(
                    comment.format(
                        # this is a bit of a hack, but we've decided to not rename files back to kubernetes-*
                        # files. while we still need to update things to reference the eks files directly, there's
                        # still a couple of places where we still need kubernetes-* files (e.g., unmigrated operators)
                        # so for now let's just assume that autotuned things will always actually have their human-managed
                        # config in eks-* files
                        regular_filename=f"{service}/{extra_info.replace('kubernetes-', 'eks-')}.yaml",
                    )
                )
                if comment
                else {}
            )
            content.update(data)
        else:
            # avoids content.update to preserve comments in `data`
            content = data

        f.write(yaml.round_trip_dump(content))
    return filename


def _commit_files(files: List[str], message: str) -> bool:
    """
    Stages the given files and creates a commit with the given message.

    Returns True if a new commit was created, False if the files are unchanged.
    """
    subprocess.check_call(["git", "add"] + files)
    # Skip commit if no changes are staged
    result_code = subprocess.call(["git", "diff-index", "--quiet", "--cached", "HEAD"])
    if result_code == 0:
        return False
    else:
        subprocess.check_call(["git", "commit", "--no-verify", "--message", message])
        return True


class PushNotFastForwardError(Exception):
    pass


class ValidationError(Exception):
    pass


def _push_to_remote(branch: str) -> None:
    try:
        subprocess.check_output(
            ("git", "push", "origin", branch), stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as e:
        if "Updates were rejected" in str(e.stdout):
            raise PushNotFastForwardError()
        else:
            log.error(f"Push to {branch} failed with:\n {e.stdout}")
            raise


def validate_auto_config_file(filepath: str, schema_subdir: str):
    basename = os.path.basename(filepath)
    for file_type in KNOWN_CONFIG_TYPES:
        schema_path = f"{schema_subdir}/{file_type}" if schema_subdir else file_type
        if basename.startswith(file_type):
            return bool(validate_schema(filepath, schema_path))
    else:
        logging.info(f"{filepath} is invalid because it has no validator defined")
        return False


class AutoConfigUpdater:
    """
    Helper class for updating automated paasta configs.

    Usage:

        updater = AutoConfigUpdater('about_me', 'git@git.me:my_configs', branch='test')
        # The context manager clones the repo into a local temp directory, then
        # cleans up afterwards.
        with updater:
            # The updater replaces the content of files, so get the existing data
            # first if you want to update it
            data = updater.get_existing_configs('service_foo', 'conf_file')
            data["new_key"] = "g_minor"

            # Now write the new data
            updater.write_configs('service_foo', 'conf_file', data)

            # Edit more files...

            # Once you're done editing files, commit. If all files pass validation,
            # the updater will commit the changes and push them to the desired branch
            # on the remote.
            updater.commit_to_remote(extra_message="Adding some extra context.")

    Raises PushNotFastForwardError if the updated branch does not include changes in the
    remote branch.
    """

    def __init__(
        self,
        config_source: str,
        git_remote: str,
        branch: str = "master",
        working_dir: Optional[str] = None,
        do_clone: bool = True,
        validation_schema_path: str = "",
    ):
        self.config_source = config_source
        self.git_remote = git_remote
        self.branch = branch
        self.working_dir = working_dir
        self.do_clone = do_clone
        self.files_changed: Set[str] = set()
        self.validation_schema_path = validation_schema_path
        self.tmp_dir = None

    def __enter__(self):
        if self.do_clone:
            self.tmp_dir = TemporaryDirectory(dir=self.working_dir)
            self.working_dir = self.tmp_dir.name
            subprocess.check_call(["git", "clone", self.git_remote, self.working_dir])
        self.pwd = os.getcwd()
        os.chdir(self.working_dir)
        if self.branch != "master":
            if self._remote_branch_exists():
                subprocess.check_call(["git", "fetch", "origin", self.branch])
                subprocess.check_call(
                    ["git", "checkout", "-b", self.branch, f"origin/{self.branch}"]
                )
            else:
                subprocess.check_call(["git", "checkout", "-b", self.branch])
        return self

    def __exit__(self, type, value, traceback):
        os.chdir(self.pwd)
        if self.tmp_dir:
            self.tmp_dir.cleanup()

    def _remote_branch_exists(self) -> bool:
        return (
            subprocess.run(
                ["git", "ls-remote", "--exit-code", "--heads", "origin", self.branch],
            ).returncode
            == 0
        )

    def write_configs(
        self,
        service: str,
        extra_info: str,
        configs: Dict[str, Any],
        sub_dir: Optional[str] = None,
        comment: Optional[str] = None,
    ):
        result = write_auto_config_data(
            service,
            extra_info,
            configs,
            soa_dir=self.working_dir,
            sub_dir=sub_dir,
            comment=comment,
        )
        if result:
            self.files_changed.add(result)

    def get_existing_configs(
        self, service: str, file_name: str, sub_dir: Optional[str] = None
    ) -> Dict[str, Any]:
        config_file_path = f"{sub_dir}/{file_name}" if sub_dir else file_name
        config_file_abs_path = os.path.join(
            os.path.abspath(self.working_dir),
            service,
            config_file_path + ".yaml",
        )
        try:
            return yaml.round_trip_load(open(config_file_abs_path))
        except FileNotFoundError:
            return {}

    def validate(self):
        return_code = True
        for filepath in self.files_changed:
            # We don't short circuit after a failure so the caller gets info on all the failures
            return_code = (
                validate_auto_config_file(filepath, self.validation_schema_path)
                and return_code
            )
        return return_code

    def commit_to_remote(self, extra_message: str = ""):
        if not self.validate():
            log.error("Files failed validation, not committing changes")
            raise ValidationError

        # TODO: more identifying information, like hostname or paasta_tools version?
        message = f"Update to configs from {self.config_source}"
        if extra_message:
            message = f"{message}\n\n{extra_message}"

        if _commit_files(list(self.files_changed), message):
            _push_to_remote(self.branch)
        else:
            log.info("No files changed, no push required.")

    def _clamp_recommendations(
        self, merged_recommendation: Dict[str, Any], config: Dict[str, Any]
    ) -> Dict[str, Any]:
        clamped_recomendation = copy.deepcopy(merged_recommendation)
        for limit_type, limits in config.get("autotune_limits", {}).items():
            log.debug(f"Processing {limit_type} autotune limits...")
            min_value = limits.get("min")
            max_value = limits.get("max")
            unclamped_resource_value = clamped_recomendation.get(limit_type)

            # no autotune data present, but min value present
            if not unclamped_resource_value and min_value:
                # use the min value since this is likely an autogenerated service where we know we have a given minimum CPU
                # that we'd like to allocate
                log.debug(
                    f"No {limit_type} autotune data found, using autotune limit lower bound ({min_value})."
                )
                clamped_recomendation[limit_type] = min_value

            # otherwise, we can do some pretty rote clamping of resource values
            elif unclamped_resource_value is not None:
                if min_value and unclamped_resource_value < min_value:
                    log.debug(
                        f"{limit_type} autotune config under configured limit ({min_value}), using autotune limit lower bound."
                    )
                    clamped_recomendation[limit_type] = min_value
                if max_value and unclamped_resource_value > max_value:
                    log.debug(
                        f"{limit_type} autotune config over configured limit ({min_value}), using autotune limit upper bound."
                    )
                    clamped_recomendation[limit_type] = max_value
            else:
                log.debug(
                    f"No {limit_type} autotune data or limits found, will continue using PaaSTA defaults."
                )

        return clamped_recomendation

    def merge_recommendations(
        self, recs: Dict[Tuple[str, str], Dict[str, Any]]
    ) -> Dict[Tuple[str, str], Dict[str, Any]]:
        """
        :param recs: Dictionary of (service, instance_type_cluster) -> recommendations.
        NOTE: instance_type_cluster is something like "kubernetes-pnw-prod".
        :returns: a dictionary of the same format, with the previous recommendations merged in and autotune_limits applied.
        """
        merged_recs = {}
        for (
            service,
            instance_type_cluster,
        ), recommendations_by_instance in recs.items():
            log.info(f"Starting to process {service}/{instance_type_cluster}.yaml...")
            log.debug(
                f"Getting current autotune configs for {service}/{instance_type_cluster}.yaml"
            )
            existing_recommendations = self.get_existing_configs(
                service=service,
                file_name=instance_type_cluster,
                sub_dir=AUTO_SOACONFIG_SUBDIR,
            )

            log.debug(
                f"Getting current configs for {service}/{instance_type_cluster}.yaml..."
            )
            # i'm so sorry.
            # basically, we need to make sure that for every autotuned service, we load both kubernetes-
            # and eks- files for the existing configs, as there are services that at any given time will
            # only exist on one of these or may have a mix (and the csv file that we get fakes the current
            # cluster type)
            # NOTE: if an instance appears in both files, the counterpart will always "win" - this
            # should only be possible while an instance is being migrated from one instance type to
            # another
            instance_type, _ = instance_type_cluster.split("-", maxsplit=1)
            existing_configs = {
                # if we upgrate to py3.9 before getting rid of this code, this should use PEP-584-style dict merging
                **self.get_existing_configs(
                    service=service,
                    file_name=instance_type_cluster,
                ),
                **self.get_existing_configs(
                    service=service,
                    file_name=instance_type_cluster.replace(
                        instance_type, INSTANCE_TYPE_COUNTERPARTS.get(instance_type, "")
                    ),
                ),
            }

            for instance_name, recommendation in recommendations_by_instance.items():
                if instance_name not in existing_configs:
                    log.debug(
                        f"Skipping config merge for {instance_name} as it no longer exists in {instance_type_cluster}.yaml."
                    )
                    continue

                log.debug(
                    f"Merging recommendations for {instance_name} in {service}/{AUTO_SOACONFIG_SUBDIR}/{instance_type_cluster}.yaml..."
                )
                existing_recommendations.setdefault(instance_name, {})
                existing_recommendations[instance_name].update(recommendation)

                existing_recommendations[instance_name] = self._clamp_recommendations(
                    merged_recommendation=existing_recommendations[instance_name],
                    config=existing_configs[instance_name],
                )
            merged_recs[(service, instance_type_cluster)] = existing_recommendations
            log.info(f"Done processing {service}/{instance_type_cluster}.yaml.")

        return merged_recs
